#include <chrono>
#include <pybind11/cast.h>
#include <pybind11/chrono.h>
#include <pybind11/functional.h>
#include <pybind11/pybind11.h>
#include <pybind11/pytypes.h>

#include "simplesync.hpp"

namespace py = pybind11;
typedef simplesync::SimpleSyncDynamic<std::chrono::time_point<std::chrono::steady_clock>,
                                      std::chrono::nanoseconds, 1024>
    SimpleS;

PYBIND11_MODULE(pysimplesync, m) {
    py::class_<SimpleS>(m, "SimpleSync")
        .def(py::init([]() {
            return std::unique_ptr<SimpleS>(new SimpleS(
                [](uint8_t *buf, unsigned int buf_size) {
                    py::print(py::bytes((char *)buf, buf_size));
                    return 0;
                },
                [](SimpleS::NumberInterface *&, char *key, int value) {
                    py::print(std::string(key), ": ", value);
                },
                nullptr, std::chrono::nanoseconds::max()));
        }))
        .def(
            "__getitem__",
            [](SimpleS &s, const std::string &key) -> py::object {
                auto ni = s.get_number_if(key.c_str());
                if (ni) return py::int_(ni->value);
                auto si = s.get_string_if(key.c_str());
                if (si) return py::str(si->value);
                throw pybind11::key_error();
            },
            py::is_operator())
        .def(
            "__getitem__",
            [](SimpleS &s, unsigned int i) -> py::tuple {
                if (i < s.number_interfaces.size()) {
                    return py::make_tuple(py::str(s.number_interfaces[i]->key),
                                          py::int_(s.number_interfaces[i]->value));
                }
                i -= s.number_interfaces.size();
                if (i < s.string_interfaces.size()) {
                    return py::make_tuple(py::str(s.string_interfaces[i]->key),
                                          py::str(s.string_interfaces[i]->value));
                }
                throw pybind11::index_error();
            },
            py::is_operator())
        .def(
            "__setitem__",
            [](SimpleS &s, const std::string &key, py::object value) {
                if (py::isinstance<py::int_>(value)) {
                    auto ni = s.get_number_if(key.c_str());
                    if (ni) {
                        ni->value          = value.cast<int>();
                        ni->send_requested = true;
                        return;
                    }
                }
                else if (py::isinstance<py::str>(value)) {
                    auto ni = s.get_string_if(key.c_str());
                    if (ni) {
                        if(value.cast<std::string>().length() > SimpleS::STR_LEN_MAX)
                          throw pybind11::buffer_error();
                        strcpy(ni->value, value.cast<std::string>().c_str());
                        ni->send_requested = true;
                        return;
                    }
                } else
                    throw pybind11::type_error();
                throw pybind11::key_error();
            },
            py::is_operator())
        .def("update", [](SimpleS &s) { s.update(std::chrono::steady_clock::now()); })
        .def("handle_stream", [](SimpleS &s, const std::string &buf_new) {
            /* handle_stream(b'\x06\x01\x74\x65\x73\x74\x04\x6d\x21\x97\x00') */
            s.handle_stream((const uint8_t *)buf_new.data(), buf_new.length());
        });
}
